//*****************************************************************************
// File:        Crust.pas
//
// Classe:      TCrust
//
// Description: TCrust is supposed to automatically extract the exterior surface
//              of a cloud of points in 3d using Delaunay. Here the extension of
//              the 2d crust extraction is used; but it doesn't really work for
//              the moment...
//
// Note:
//
//*****************************************************************************
// Last version: 25-06-2002
//*****************************************************************************

unit Crust;

interface

uses
  Mesh,
  Tetra,
  Point,
  Vector,
  Triangle,
  Classes,
  GeomTools;

type
  TCrust = class
  private
    m_mesh:     TMesh;
    procedure   ComputeCrust(lstTr: TList);
  public
    constructor Create(mesh: TMesh);
    procedure   ExtractCrust;
  end;


implementation

{ TIsosurface }

constructor TCrust.Create(mesh: TMesh);
begin
  m_mesh := mesh;
end;

procedure TCrust.ExtractCrust;
var
  lstTr: TList;
  i: integer;
begin
  lstTr := TList.Create;
  self.ComputeCrust(lstTr);
  m_mesh.bDrawCrust := true;
  m_mesh.SetCrust(lstTr);

  //-- free the list of triangles
  for i := 0 to (lstTr.Count - 1) do
  begin
    TTriangle(lstTr[i]).Free;
  end;
  lstTr.Free;
end;


procedure TCrust.ComputeCrust(lstTr: TList);
var
  i: integer;
  j: integer;
  te: TTetra;
  adj: TTetra;
  lstPts: TList;
  lstTe: TList;
  centre: TPoint;
  centreAdj: TPoint;
  n: TVector;
begin
  centre := TPoint.Create;
  centreAdj := TPoint.Create;
  n := TVector.Create;
  lstPts := m_mesh.GetListPts;
  lstTe := m_mesh.GetListTetra;
  for i := 0 to (lstTe.Count - 1) do  //-- for each tetrahedron in the mesh
  begin
    te := lstTe[i];
    //-- if the tetra has one point from the big tetra, ignore it
    if not ( (te.GetP(1) = lstPts[0]) or
             (te.GetP(1) = lstPts[1]) or
             (te.GetP(1) = lstPts[2]) or
             (te.GetP(1) = lstPts[3]) or
             (te.GetP(2) = lstPts[0]) or
             (te.GetP(2) = lstPts[1]) or
             (te.GetP(2) = lstPts[2]) or
             (te.GetP(2) = lstPts[3]) or
             (te.GetP(3) = lstPts[0]) or
             (te.GetP(3) = lstPts[1]) or
             (te.GetP(3) = lstPts[2]) or
             (te.GetP(3) = lstPts[3]) or
             (te.GetP(4) = lstPts[0]) or
             (te.GetP(4) = lstPts[1]) or
             (te.GetP(4) = lstPts[2]) or
             (te.GetP(4) = lstPts[3]) ) then
    begin
      te.GetCircumCentre(centre);
      //-- test the 4 faces of the triangle
      for j := 1 to 4 do
      begin
        adj := te.SetIndex(j).GetTrot(0);
        adj.GetCircumCentre(centreAdj);
        //-- if the voronoi vertex of the adjacent tetra is not inside the
        //-- sphere formed by the face and the centre of the tetra, then the
        //-- face is part of the crust.
        if not (TGeomTools.InSphere(te.GetProt(1),
                                    te.GetProt(2),
                                    te.GetProt(3),
                                    centre,
                                    centreAdj)) then
        begin
          TGeomTools.NormalOppositePt(te.GetProt(1),
                                      te.GetProt(2),
                                      te.GetProt(3),
                                      lstPts[0],
                                      n);
          lstTr.Add(TTriangle.Create(te.GetProt(1),
                                     te.GetProt(2),
                                     te.GetProt(3),
                                     n));
        end;
      end;
    end;
  end;
  centre.Free;
  centreAdj.Free;
  n.Free;
end;



end.
