//*****************************************************************************
// File:        DrawTools.pas
//
// Classe:      TDrawTools
//
// Description: TDrawTools contains static methods for drawing primitives - such
//              as points, lines, triangles, tetras, etc. - with OpenGL.
//
// Note:        -
//
//*****************************************************************************
// Last version: 23-08-2005
//*****************************************************************************

unit DrawTools;

interface

uses
  Point,
  Tetra,
  Triangle,
  Vector,
  OpenGL,
  GeomTools,
  Classes,
  SysUtils;

type
  TDrawTools = class
  public
    class procedure DrawNode(pt: TPoint; r, g, b: double; hl: integer = 0);
    class procedure DrawLine(v1, v2: TPoint; r, g, b: double; width: integer);
    class procedure DrawTriangleFace_facingPT(v1, v2, v3, pt: TPoint; r, g, b, alpha: double);
    class procedure DrawTriangleFace(v1, v2, v3, pt: TPoint; r, g, b, alpha: double); overload;
    class procedure DrawTriangleFace(tr: TTriangle; r, g, b, alpha: double); overload;
    class procedure DrawTriangleFrame(tr: TTriangle; r, g, b: double; width: integer); overload;
    class procedure DrawTriangleFrame(v1, v2, v3: TPoint; r, g, b: double; width: integer); overload;
    class procedure DrawTetraFrame(te: TTetra; r, g, b: double; width: integer);
    class procedure DrawTetraFaces(te: TTetra; r, g, b, alpha: double);
    class procedure DrawPolygonFrame(lstPts: TList; r, g, b: double; width: integer);
    class procedure DrawPolygonFace(lstPts: TList; pt: TPoint; r, g, b, alpha: double);
    class procedure DrawQuad(v1, v2, v3, v4: TPoint; r, g, b, alpha: double);
    class procedure SetNodeRadius(r: double);
  end;



implementation

//-- 'class' static field
var
  radiusNode: double = 0.01;


//******************************************************************************
// Description: To draw a point with a specific color and size
//
// Input:       hl: means highlight. if it's 1 the point is drawn 1.5 times
//              bigger than the other ones.
//
// Output:      -
//******************************************************************************
class procedure TDrawTools.DrawNode(pt: TPoint; r, g, b: double; hl: integer);
var
  quad: GLUquadricObj;
begin
  //-- draw a sphere
  quad := gluNewQuadric;
  glPushMatrix;
  glTranslatef(pt.x, pt.y, pt.z);
  glColor3f(r, g, b);
  gluQuadricDrawStyle(quad, GLU_FILL);
  if (hl = 0) then
    gluSphere(quad, radiusNode, 15, 15)
  else if (hl = -1) then
    gluSphere(quad, radiusNode * 0.4, 15, 15)
  else
    gluSphere(quad, radiusNode * 1.5, 15, 15);
  glPopMatrix;
  gluDeleteQuadric(quad);
end;


//******************************************************************************
// Description: To draw a triangular face
//
// Input:       v1, v2, v3: the 3 vertices of the face
//              pt:         vertex in the opposite direction of the normal
//              r, g, b,:   colour of the face
//              bigger than the other ones.
//******************************************************************************
class procedure TDrawTools.DrawTriangleFace(v1, v2, v3, pt: TPoint;
                                            r, g, b, alpha: double);
var
  n: TVector;
begin
  n := TVector.Create;
  glBegin(GL_TRIANGLES);
    TGeomTools.NormalOppositePt(v1, v2, v3, pt, n);
    glColor4d(r, g, b, alpha);
    glNormal3f(n.coord.x, n.coord.y, n.coord.z);
    glVertex3f(v1.x, v1.y, v1.z);
    glVertex3f(v2.x, v2.y, v2.z);
    glVertex3f(v3.x, v3.y, v3.z);
  glEnd;
  n.Free;
end;


class procedure TDrawTools.DrawTriangleFace(tr: TTriangle; r, g, b,
  alpha: double);
var
  n: TVector;
begin
  n := tr.GetNormale;
  glBegin(GL_TRIANGLES);
    glColor4d(r, g, b, alpha);
    glNormal3f(n.coord.x, n.coord.y, n.coord.z);
    glVertex3f(tr.GetPoint(1).x, tr.GetPoint(1).y, tr.GetPoint(1).z);
    glVertex3f(tr.GetPoint(2).x, tr.GetPoint(2).y, tr.GetPoint(2).z);
    glVertex3f(tr.GetPoint(3).x, tr.GetPoint(3).y, tr.GetPoint(3).z);
  glEnd;
end;


//******************************************************************************
// Description: Draw the 'frame' of a tetra (only the edges).
//
// Input:       width: the width of each edge
//
// Output:      -
//******************************************************************************
class procedure TDrawTools.DrawTetraFrame(te: TTetra; r, g, b:
                                          double; width: integer);
var
  i: integer;
  j: integer;
begin
  for i := 1 to 4 do
  begin
    for j := (i + 1) to 4 do
    begin
      DrawLine(te.GetP(i), te.GetP(j), r, g, b, width);
    end;
  end;
end;


class procedure TDrawTools.DrawLine(v1, v2: TPoint; r, g, b: double;
                                    width: integer);
var
  lineWidth: GLfloat;
begin
  glColor3f(r, g, b);
  glGetFloatv(GL_LINE_WIDTH, @lineWidth);
  glLineWidth(width);
  glBegin(GL_LINES);
    glVertex3f(v1.x, v1.y, v1.z);
    glVertex3f(v2.x, v2.y, v2.z);
  glEnd;
  glLineWidth(lineWidth);
end;


class procedure TDrawTools.DrawTetraFaces(te: TTetra; r, g, b, alpha: double);
var
  i: integer;
begin
  for i := 1 to 4 do
  begin
    te.SetIndex(i);
    if ((i mod 2) = 0) then //-- switch position 2 and 3
    begin
      DrawTriangleFace(te.GetProt(0), te.GetProt(2), te.GetProt(1),
                        te.GetProt(3), r, g, b, alpha);
    end
    else //-- ( (i = 1) or (i = 3) )
    begin
      DrawTriangleFace(te.GetProt(0), te.GetProt(1), te.GetProt(2),
                        te.GetProt(3), r, g, b, alpha);
    end;
  end;
end;


class procedure TDrawTools.DrawPolygonFace(lstPts: TList;  pt: TPoint;
  r, g, b, alpha: double);
const
  tolerance: double = 0.0000000000001;
var
  n: TVector;
  i: integer;
  pt0: TPoint;
  pt1: TPoint;
  pt2: TPoint;
  pt3: TPoint;
begin
  n := TVector.Create;
  try
    TGeomTools.NormalOppositePt(lstPts[0], lstPts[1], lstPts[2], pt, n);
    glNormal3f(n.coord.x, n.coord.y, n.coord.z);
    //-- draw face
    glBegin(GL_POLYGON);
      glColor4d(r, g, b, alpha);
      for i := 0 to (lstPts.Count - 1) do
      begin
        pt0 := lstPts[i];
        glVertex3f(pt0.x, pt0.y, pt0.z);
      end;
    glEnd;
  except
    on E: EZeroDivide do    { TODO : drawing Voronoi cell and exception shit }
    begin
      //-- 2 points are the same so we cannot compute a normale to the plane
      //-- find 3 different points to compute the normale of the plane
      //-- is there are only 2 diff points, don't draw anything
{      pt1 := lstPts[0];
      pt2 := nil;
      pt3 := nil;
      for i := 1 to (lstPts.Count - 1) do
      begin
        pt0 := lstPts[i];
        if (pt1.Hit(pt0, tolerance) = false) then
        begin
          pt2 := pt0;
        end;
      end;
      if (pt2 <> nil) then
      begin
        for i := 1 to (lstPts.Count - 1) do
        begin
          pt0 := lstPts[i];
          if ( (pt1.Hit(pt0, tolerance) = false) and (pt2.Hit(pt0, tolerance) = false) ) then
          begin
            pt3 := pt0;
          end;
        end;
      end;

      if ( (pt2 <> nil) and (pt3 <> nil) ) then
      begin
        TGeomTools.NormalOppositePt(pt1, pt2, pt3, pt, n);
        glNormal3f(n.coord.x, n.coord.y, n.coord.z);
        //-- draw face
        glBegin(GL_POLYGON);
          glColor4d(r, g, b, alpha);
          for i := 0 to (lstPts.Count - 1) do
          begin
            pt0 := lstPts[i];
            glVertex3f(pt0.x, pt0.y, pt0.z);
          end;
        glEnd;
      end;
}    end; 
  end;
  n.Free;
end;


class procedure TDrawTools.DrawPolygonFrame(lstPts: TList; r, g, b: double;
  width: integer);
var
  i: integer;
begin
  //-- draw contour lines
  for i := 0 to (lstPts.Count - 2) do
  begin
    DrawLine(lstPts[i], lstPts[i + 1], r, g, b, width);
  end;
  DrawLine(lstPts[0], lstPts[lstPts.Count - 1], r, g, b, width);
end;


class procedure TDrawTools.DrawTriangleFrame(tr: TTriangle; r, g,
  b: double; width: integer);
begin
    DrawLine(tr.GetPoint(1), tr.GetPoint(2), r, g, b, width);
    DrawLine(tr.GetPoint(2), tr.GetPoint(3), r, g, b, width);
    DrawLine(tr.GetPoint(3), tr.GetPoint(1), r, g, b, width);
end;


class procedure TDrawTools.DrawTriangleFrame(v1, v2, v3: TPoint; r, g,
  b: double; width: integer);
begin
  DrawLine(v1, v2, r, g, b, width);
  DrawLine(v2, v3, r, g, b, width);
  DrawLine(v3, v1, r, g, b, width);
end;

//******************************************************************************
// Description: Set the size (the radius) of the nodes drawn
//
// Input:       r: the radius
//
// Output:      -
//
// Notes:       radiusNode is a 'form variable' whose value is 0.02 at creation
//******************************************************************************
class procedure TDrawTools.SetNodeRadius(r: double);
begin
  radiusNode := r;
end;


class procedure TDrawTools.DrawQuad(v1, v2, v3, v4: TPoint; r, g, b,
  alpha: double);
begin
  glBegin(GL_QUADS);
    glColor4d(r, g, b, alpha);
    glNormal3f(0, 0, 0);  //-- no illumination is used here, better output
    glVertex3f(v1.x, v1.y, v1.z);
    glVertex3f(v2.x, v2.y, v2.z);
    glVertex3f(v3.x, v3.y, v3.z);
    glVertex3f(v4.x, v4.y, v4.z);
  glEnd;
end;


//******************************************************************************
// Description: To draw a triangular face
//
// Input:       v1, v2, v3: the 3 vertices of the face
//              pt:         vertex in the same direction as the normal
//              r, g, b,:   colour of the face
//              bigger than the other ones.
//******************************************************************************
class procedure TDrawTools.DrawTriangleFace_facingPT(v1, v2, v3,
  pt: TPoint; r, g, b, alpha: double);
var
  n: TVector;
begin
  n := TVector.Create;
  glBegin(GL_TRIANGLES);
    TGeomTools.NormalDirectionPt(v1, v2, v3, pt, n);
    glColor4d(r, g, b, alpha);
    glNormal3f(n.coord.x, n.coord.y, n.coord.z);
    glVertex3f(v1.x, v1.y, v1.z);
    glVertex3f(v2.x, v2.y, v2.z);
    glVertex3f(v3.x, v3.y, v3.z);
  glEnd;
  n.Free;
end;



end.
