//*****************************************************************************
// File:        Ear.pas
//
// Classe:      TEar
//
// Description: TEar represents an ear of a polyhedron. Actually, only 2-ears
//              are stored in the program; 3-ear are detected 'on-the-fly'.
//              An ear is defined by the 2 tetra underlying, and 4 neighbours.
//*****************************************************************************
// Last version: 2004-06-30
//*****************************************************************************

unit Ear;

{$IFDEF DEBUG}
  {$ASSERTIONS ON}
{$ELSEIF True}
  {$ASSERTIONS OFF}
{$IFEND}

interface

uses
  Tetra,
  Point,
  Dialogs, SysUtils,
  Classes;

type

  TEar = class
  private
    m_te:       array[1..2] of TTetra;
    m_adj:      array[1..4] of TEar;
    procedure   TestInvariants;
  public
    m_no:       integer;
    m_flag:     boolean;
    constructor Create(te1, te2: TTetra; v: TPoint);
    function    GetAdj(i: integer) : TEar;
    procedure   SetAdj(i: integer; ear: TEar);
    function    GetTetra(i: integer) : TTetra;
    procedure   SetTetra(i: integer; const te: TTetra);
    function    GetVertex : TPoint;

    function    Compare(const te1, te2: TTetra) : boolean;
    function    ContainTetra(const te: TTetra) : boolean;
    function    Is3ear(out adjEar1, adjEar2: TEar) : boolean;

    function    IsConvex : integer;
    function    IsFlippable : integer;
    function    IsDelaunay(const lstPts: TList) : boolean;
  end;


  //****************************************************************************
  //  Earp is a TEar with the attribute power attached to it; this is used for
  //  DeleteVertex_power. The power of an ear is actually the power of the
  //  deleted vertex w/r to the circumsphere of the ear (an imaginary tetra)
  //****************************************************************************
  TEarp = class(TEar)
  private
    m_power:    double;
    procedure   ComputePower;
  public
    constructor Create(te1, te2: TTetra; v: TPoint);
    function    GetPower : double;
    procedure   UpdatePower;
    procedure   SetPower(pow: double);
  end;


implementation

uses
  GeomTools;

//-----------------------
//-- 'class' static field
//-----------------------
var
  m_v: TPoint = nil; //-- the vertex to be deleted in DeleteVertex()
  no: integer = 0;


{ TEar }

constructor TEar.Create(te1, te2: TTetra; v: TPoint);
begin
  m_v := v;
  Assert( (te1 <> nil) and (te2 <> nil) );
  Assert(te1 <> te2);
  m_te[1] := te1;
  m_te[2] := te2;
  m_adj[1] := nil;
  m_adj[2] := nil;
  m_adj[3] := nil;
  m_adj[4] := nil;
  m_no := no;
  m_flag := false;
  inc(no);
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
end;


function TEar.GetAdj(i: integer): TEar;
begin
  Assert( (i >= 1) and (i <= 4) );
  Result := m_adj[i];
end;


procedure TEar.SetAdj(i: integer; ear: TEar);
begin
  Assert( (i >= 1) and (i <= 4) );
  Assert(ear <> nil);
  m_adj[i] := ear;
end;


function TEar.GetTetra(i: integer): TTetra;
begin
  Assert( (i >= 1) and (i <= 2) );
  Result := m_te[i];
end;


procedure TEar.SetTetra(i: integer; const te: TTetra);
begin
  Assert( (i >= 1) and (i <= 2) );
  Assert(te <> nil);
  m_te[i] := te;
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
end;


//******************************************************************************
// Description: Determines if the ear is CONVEX. An ear is convex with respect
//              to the deleted point in the middle of H.
//
// Input:       -
//
// Output:      1  : convex (w/r to deleted point)
//              0  : flat ear
//              -1 : concave (w/r to deleted point)
//******************************************************************************
function TEar.IsConvex: integer;
var
  i: integer;
  j: integer;
begin
  m_te[1].SetIndex(m_v);
  m_te[2].SetIndex(m_te[1]);
  i := TGeomTools.Orient3D(m_te[1].GetProt(1), m_te[1].GetProt(2),
                           m_te[1].GetProt(3), m_v);
  j := TGeomTools.Orient3D(m_te[1].GetProt(1), m_te[1].GetProt(2),
                           m_te[1].GetProt(3), m_te[2].GetProt(0));

  //-- check if the first ear is formed by a flat tetra, if it's the case then
  //-- switch the ear & the problem is solved; unless the 2 ears are formed by
  //-- flat tetra.
  if ( (i = 0) or (j = 0) ) then
  begin
    m_te[2].SetIndex(m_v);
    m_te[1].SetIndex(m_te[2]);
    i := TGeomTools.Orient3D(m_te[2].GetProt(1), m_te[2].GetProt(2),
                             m_te[2].GetProt(3), m_v);
    j := TGeomTools.Orient3D(m_te[2].GetProt(1), m_te[2].GetProt(2),
                             m_te[2].GetProt(3), m_te[1].GetProt(0));
  end;

  if ( (i = 0) or (j = 0) ) then
  begin
    Result := 0;
  end
  else
  begin
    if (i = j) then
      Result := 1
    else
      Result := -1;
  end;
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
end;


//******************************************************************************
// Description: Determines if the ear is DELAUNAY with respect to the other
//              vertices forming H. When deleting a vertex from a mesh, it
//              suffices to test the Delaunayhood of the ear with simply these
//              vertices.
//
// Input:       lstPts : the list of TPoint forming the polyhedron H.
//
// Output:      TRUE  -> the ear is Delaunay
//              FALSE -> the ear is NOT Delaunay
//
// Note:        Only DeleteVertex_insphere needs that function.
//              DeleteVertex_power uses the power to test if an ear is Delaunay.
//******************************************************************************
function TEar.IsDelaunay(const lstPts: TList): boolean;
var
  i: integer;
  r: boolean;
  tempt: TPoint;
  tempTe: TTetra;
  circum: TPoint;
  radius, distance: double;
begin
  r := true;
  m_te[1].SetIndex(m_v);
  m_te[2].SetIndex(m_te[1]);
  if (TGeomTools.Orient3D(m_te[1].GetProt(1), m_te[1].GetProt(2),
                          m_te[1].GetProt(3), m_te[2].GetProt(0)) = 1) then
  begin
    tempTe := TTetra.Create(m_te[1].GetProt(1), m_te[1].GetProt(2),
                            m_te[1].GetProt(3), m_te[2].GetProt(0));
  end
  else
  begin
    tempTe := TTetra.Create(m_te[1].GetProt(2), m_te[1].GetProt(1),
                            m_te[1].GetProt(3), m_te[2].GetProt(0));
  end;

  for i := 0 to (lstPts.Count - 1) do
  begin
    tempt := lstPts[i];
    if (tempTe.HasPoint(tempt) = false) then
    begin
    {$IFDEF PERTURB}
      if (TGeomTools.InSphere_perturb(tempTe, tempt) = true) then
    {$ELSEIF True}
      if (TGeomTools.InSphere(tempTe, tempt) = true) then
    {$IFEND}
      begin
{        circum := TPoint.Create;
        TGeomTools.CircumSphere(tempTe, circum);
        radius := TGeomTools.Distance3d(circum, tempTe.GetP(1));
        distance := TGeomTools.Distance3d(circum, tempt);
        ShowMessage(FloatToStr(radius - distance));
}        r := false;
        break;
      end;
    end;
  end;
  tempTe.Free;
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
  Result := r;
end;


//******************************************************************************
// Description: Determines if the ear can be flipped; it actually returns the
//              result of the intersection test of the ear with the common
//              facet.
//
// Input:       -
//
// Output:      -1 : no intersection
//               0 : 4 points are coplanar
//               1 : yes intersection
//******************************************************************************
function TEar.IsFlippable: integer;
var
  p: TPoint;
  q: TPoint;
begin
  p := m_te[1].SetIndex(m_te[2]).GetProt(0);
  q := m_te[2].SetIndex(m_te[1]).GetProt(0);
  Result := TGeomTools.IntersectionEdgeTriangle(p, q,
                                                m_te[1].GetProt(1),
                                                m_te[1].GetProt(2),
                                                m_te[1].GetProt(3));
end;


//******************************************************************************
// Description: Compares the ear with another 2-ear formed by 2 tetra. 2 ears
//              'the same' if they are formed by the 2 same tetra, no matter
//              what the order is; the 4 pointers to adjacent ears doesn't
//              matter either.
//
// Input:       te1, te2 : the 2 TTetra forming the ear to compare against
//
// Output:      TRUE  -> ear & te1-te2 are the same
//              FALSE -> NOT the same
//******************************************************************************
function TEar.Compare(const te1, te2: TTetra): boolean;
begin
  if ( ((te1 = m_te[1]) and (te2 = m_te[2])) or
       ((te1 = m_te[2]) and (te2 = m_te[1])) ) then
  begin
    Result := true;
  end
  else
  begin
    Result := false;
  end;
end;


function TEar.GetVertex: TPoint;
begin
  Result := m_v;
end;


function TEar.Is3ear(out adjEar1, adjEar2: TEar): boolean;
var
  te, adj: TTetra;
  commonPt: TPoint;
  i: integer;
  temp: TTetra;
  re: boolean;
  tempear: TEar;
begin
  te := m_te[1];
  adj := m_te[2];
  commonPt := adj.SetIndex(te).GetProt(0);
  te.SetIndex(adj);
  re := false;
  temp := nil;
  for i := 1 to 3 do
  begin
    if (te.GetProt(i) <> m_v) then
    begin
      temp := te.GetTrot(i);
      if (temp <> nil) then
      begin
        if (temp.HasPoint(commonPt) = true) then
        begin
          re := true;
          break;
        end;
      end;
    end;
  end;
  if (re = true) then
  begin
    adjEar1 := nil;
    adjEar2 := nil;
    for i := 1 to 4 do
    begin
      tempear := m_adj[i];
      if (tempear.ContainTetra(temp) = true) then
      begin
        if (adjEar1 = nil) then
          adjEar1 := tempear
        else
          adjEar2 := tempear;
      end;
    end;
    Assert(adjEar1 <> nil);
    Assert(adjEar2 <> nil);
  end;
  Result := re;
end;


function TEar.ContainTetra(const te: TTetra): boolean;
begin
  if ( (te = m_te[1]) or (te = m_te[2]) ) then
  begin
    Result := true;
  end
  else
  begin
    Result := false;
  end;
end;



//=============================================================================
// TEarp
//=============================================================================

constructor TEarp.Create(te1, te2: TTetra; v: TPoint);
begin
  inherited;
  self.ComputePower;
end;

//******************************************************************************
// Description: The power of an ear is the power of the deleted vertex w/r to
//              the circumsphere of the ear (a tetrahedron).
//              pow(m_v, sphere(ear))
//              If the ear is not convex, then power is set to -9999 (smallest
//              possible number), since we're looking for highest power.
//
// Input:       (the TEarp)
//
// Output:      m_power for the TEarp is updated
//******************************************************************************
procedure TEarp.ComputePower;
var
  i1, i2: integer;
begin
  if (self.IsConvex = 1) then
  begin
    i1 := m_te[1].GetIndex;
    i2 := m_te[2].GetIndex;
    m_te[1].SetIndex(m_v);
    m_te[2].SetIndex(m_te[1]);

    m_power := TGeomTools.Power(m_te[1].GetProt(1), m_te[1].GetProt(2),
                                m_te[1].GetProt(3), m_te[2].GetProt(0), m_v);
    m_te[1].SetIndex(i1);
    m_te[2].SetIndex(i2);
  end
  else
  begin
    m_power := -9999;
  end;
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
end;


function TEarp.GetPower: double;
begin
  Result := m_power;
end;


procedure TEarp.UpdatePower;
begin
  self.ComputePower;
end;


procedure TEar.TestInvariants;
begin
  Assert(m_te[1] <> nil);
  Assert(m_te[2] <> nil);
end;


procedure TEarp.SetPower(pow: double);
begin
  m_power := pow;
end;


end.
