//*****************************************************************************
// File:        GeomTools.pas
//
// Classe:      TGeomTools
//
// Description: TGeomTools contains static methods for geometric computation in
//              3d. Orient3d() and InSphere() use robust arithmetic by default,
//              if normal floating-point arithmetic (with a tolerance) is wanted
//              then 'FLOAT' should be a conditional define for the compilation.
//
// Note:        Shewchuk's predicates are used for robust arithmetic. They are
//              compiled in the 'predicates.dll' and 'predicates.lib', both
//              are required to compile the program.
//
//*****************************************************************************
// Last version: 2004-07-29
//*****************************************************************************

unit GeomTools;

interface

uses
  SysUtils,
  Dialogs,
  Point,
  Vector,
  Classes,
  Math,
  Tetra;

type
  TGeomTools = class
  public
    class function  Det3x3(ax, ay, az, bx, by, bz, cx, cy, cz: double) : double; overload;
    class function  Det3x3(const a, b, c: TPoint) : double; overload;
    class function  Det4x4(a1, a2, a3, a4, b1, b2, b3, b4, c1, c2, c3, c4, d1, d2, d3, d4: double) : double;
    class function  Det4x4t(const a, b, c, d: TPoint) : double;
    class function  Det5x5_perturb(const a, b, c, d, p: TPoint; a2, b2, c2, d2, p2: double) : double;

    class function  Orient3D(const a, b, c, pt: TPoint) : integer;
    class function  Orient3D_value(const a, b, c, pt: TPoint) : double;
    class function  InSphere(const a, b, c, d, pt: TPoint) : boolean; overload;
    class function  InSphere(const te: TTetra; const pt: TPoint) : boolean; overload;
    class function  InSphere_perturb(const a, b, c, d, p: TPoint) : boolean; overload;
    class function  InSphere_perturb(const te: TTetra; const pt: TPoint) : boolean; overload;
    class function  InSphere_value(const a, b, c, d, pt: TPoint) : double;
    class procedure CircumSphere(const a, b, c, d: TPoint; out centre: TPoint); overload;
    class procedure CircumSphere(const te: TTetra; out centre: TPoint); overload;

    class function  IntersectionEdgeTriangle(const a, b, t1, t2, t3: TPoint) : integer;
    class function  IntersectionLineSphere(const p1, p2: TPoint; const te: TTetra; out t1, t2: double) : boolean; overload;
    class function  IntersectionLineSphere(const p1, p2, pt1, pt2, pt3, pt4: TPoint; out t1, t2: double) : boolean; overload;
    class function  IntersectionLineSphere_projCentre(const p1, p2: TPoint; const te: TTetra) : boolean;
    class function  Distance3d(const a, b: TPoint) : double;
    class procedure NormalOppositePt(const a, b, c, pt: TPoint; out n: TVector);
    class procedure NormalDirectionPt(const a, b, c, pt: TPoint; out n: TVector);
    class function  Power(const a, b, c, d, v: TPoint) : double;

    class procedure SetCountCallToZero;
    class function  GetCountCallOrient : integer;
    class function  GetCountCallInSphere : integer;
  end;




implementation

var
  numcall_insphere  : integer = 0;
  numcall_orient    : integer = 0;

//-- tolerance because of floating-point precision
const
  tolerance: double = 1e-12;


//-- definition of the functions of the DLL ------------------------------------
function Orient3dR(ax, ay, az, bx, by, bz, cx, cy, cz, dx, dy, dz: double): double;
  stdcall; external 'PREDICATES.DLL' name 'Orient3dR';

function InSphereR(ax, ay, az, bx, by, bz, cx, cy, cz, dx, dy, dz, ex, ey, ez: double): double;
  stdcall; external 'PREDICATES.DLL' name 'InSphereR';
//-- definition of the functions of the DLL ------------------------------------


class procedure TGeomTools.SetCountCallToZero;
begin
  numcall_insphere := 0;
  numcall_orient := 0;
end;

class function TGeomTools.GetCountCallOrient : integer;
begin
  Result := numcall_orient;
end;

class function TGeomTools.GetCountCallInSphere : integer;
begin
  Result := numcall_insphere;
end;


//******************************************************************************
// Description: Compute the determinant of 3 point
//
// Input:       coord (TPoint) of the 3 points (anti-clockwise order)
//
// Output:      the determinant of the 3 points
//
// Note:        the value of the determinant is twice the area of the triangle
//              defined by the 3 points in 3d space. Or 6 times the volume of
//              the tetra formed by the triangular face and the origin (0, 0, 0)
//******************************************************************************
class function TGeomTools.Det3x3(const a, b, c: TPoint): double;
begin
  Result := Det3x3(a.x, a.y, a.z, b.x, b.y, b.z, c.x, c.y, c.z);
end;


//******************************************************************************
// Description: Compute the determinant of 3 points (x-y-z)
//
// Input:       coord (double) of the 3 points (anti-clockwise order)
//
// Output:      the determinant of the 3 points
//
// Note:        the value of the determinant is twice the area of the triangle
//              defined by the 3 points in 3d space. Or 6 times the volume of
//              the tetra formed by the triangular face and the origin (0, 0, 0)
//******************************************************************************
class function TGeomTools.Det3x3(ax, ay, az, bx, by, bz, cx, cy, cz: double): double;
var
  temp1: double;
  temp2: double;
  temp3: double;
begin
  temp1 := ax * (by * cz - bz * cy);
  temp2 := ay * (bx * cz - bz * cx);
  temp3 := az * (bx * cy - by * cx);
  Result := temp1 - temp2 + temp3;
end;


//******************************************************************************
// Description: Compute the determinant of 4 points with homogenous coord.
//              Translation of -d to reduce to Det3x3.
//
// Input:       coord (TPoint) of the 4 points (left-hand rule order)
//
// Output:      the determinant of the 4 points (6 times the volumes of tetra)
//******************************************************************************
class function TGeomTools.Det4x4t(const a, b, c, d: TPoint): double;
begin
  Result := Det3x3(a.x - d.x, a.y - d.y, a.z - d.z,
                   b.x - d.x, b.y - d.y, b.z - d.z,
                   c.x - d.x, c.y - d.y, c.z - d.z);
end;


//******************************************************************************
// Description: Compute the determinant of 4 points.
//
// Input:       coord (TPoint) of the 4 points (left-hand rule order)
//
// Output:      the value (a double)
//******************************************************************************
class function  TGeomTools.Det4x4(a1, a2, a3, a4, b1, b2, b3, b4, c1, c2, c3, c4, d1, d2, d3, d4: double) : double;
begin
  a1 := a1 * Det3x3(b2, b3, b4, c2, c3, c4, d2, d3, d4);
  a2 := a2 * Det3x3(b1, b3, b4, c1, c3, c4, d1, d3, d4);
  a3 := a3 * Det3x3(b1, b2, b4, c1, c2, c4, d1, d2, d4);
  a4 := a4 * Det3x3(b1, b2, b3, c1, c2, c3, d1, d2, d3);

  Result := a1 - a2 + a3 - a4;
end;



//******************************************************************************
// Description: Calculate the centre of the sphere formed by the 4 vertices of
//              a tetrahedron
//
// Input:       te: the tetrahedron (TTetra), which contains 4 vertices
//              centre : the result
//
// Output:      the centre (a TPoint)
//
// Note:        source: [FAQ - comp.graphics.algorithms (5.21)]
//******************************************************************************
class procedure TGeomTools.CircumSphere(const te: TTetra; out centre: TPoint);
begin
  CircumSphere(te.GetP(1), te.GetP(2), te.GetP(3), te.GetP(4), centre);
end;


//******************************************************************************
// Description: Calculate the centre of the sphere formed by the 4 vertices of
//              a tetrahedron
//
// Input:       a,b,c,d : the 4 vertices of the tetrahedron
//              centre : the result
//
// Output:      the centre (a TPoint)
//
// Note:        source: [FAQ - comp.graphics.algorithms (5.21)]
//******************************************************************************
class procedure TGeomTools.CircumSphere(const a, b, c, d: TPoint; out centre: TPoint);
var
  temp1: TVector;
  temp2: TVector;
  temp3: TVector;
  part1: TVector;
  part2: TVector;
  part3: TVector;
  norm: double;
  pt1: TPoint;
  pt2: TPoint;
  pt3: TPoint;
begin
  temp1 := TVector.Create(a, b);
  temp2 := TVector.Create(a, c);
  temp3 := TVector.Create(a);
  part1 := TVector.Create;
  TVector.CrossProduct(temp1, temp2, part1);
  temp1.coord := d;
  temp1.Subtract(temp3);
  norm := Sqr(temp1.GetLength);
  part1.Multiply(norm);

  temp1.coord := d;
  temp2.coord := b;
  temp1.Subtract(temp3);
  temp2.Subtract(temp3);
  part2 := TVector.Create;
  TVector.CrossProduct(temp1, temp2, part2);
  temp1.coord := c;
  temp1.Subtract(temp3);
  norm := Sqr(temp1.GetLength);
  part2.Multiply(norm);

  temp1.coord := c;
  temp2.coord := d;
  temp1.Subtract(temp3);
  temp2.Subtract(temp3);
  part3 := TVector.Create;
  TVector.CrossProduct(temp1, temp2, part3);
  temp1.coord := b;
  temp1.Subtract(temp3);
  norm := Sqr(temp1.GetLength);
  part3.Multiply(norm);

  pt1 := TPoint.Create(b.x - a.x, b.y - a.y, b.z - a.z);
  pt2 := TPoint.Create(c.x - a.x, c.y - a.y, c.z - a.z);
  pt3 := TPoint.Create(d.x - a.x, d.y - a.y, d.z - a.z);

  part1.Add(part2);
  part1.Add(part3);

  part1.Multiply(1 / (2 * Det3x3(pt1, pt2, pt3)));
  temp1.coord := a;
  part1.Add(temp1);

  centre.x := part1.coord.x;
  centre.y := part1.coord.y;
  centre.z := part1.coord.z;

  temp1.Free;
  temp2.Free;
  temp3.Free;
  part1.Free;
  part2.Free;
  part3.Free;
  pt1.Free;
  pt2.Free;
  pt3.Free;
end;


//******************************************************************************
// Description: Determine if a point pt is inside the circumSphere around the
//              tetrahedron formed by a-b-c-d
//
// Input:       te : the tetrahedron, which contains 4 vertices
//              pt : the point to test
//
// Output:      TRUE  -> pt is INSIDE the sphere
//              FALSE -> pt is OUTSIDE or ON the sphere
//
// Note:        That's an Orient test made in 4D space with the points lifted
//              onto the 4D paraboloid [x4 = x1^2 + x2^2 + x3^2]. So simply
//              a Det5x5 of the points with homogenous coords.
//
//              *** IMPORTANT ***
//               This assumes that the 4 points are ordered correctly according
//               to the left-hand rule, i.e. Orient3d(a,b,c,d) must return TRUE.
//******************************************************************************
class function TGeomTools.InSphere(const te: TTetra; const pt: TPoint): boolean;
begin
  Result := InSphere(te.GetP(1), te.GetP(2), te.GetP(3), te.GetP(4), pt);
end;


//******************************************************************************
// Description: Determine if a point pt is inside the circumSphere around the
//              tetrahedron formed by a-b-c-d
//
// Input:       a,b,c,d : the 4 vertices of the tetrahedron
//              pt : the point to test
//
// Output:      TRUE  -> pt is INSIDE the sphere
//              FALSE -> pt is OUTSIDE or ON the sphere
//
// Note:        That's an Orient test made in 4D space with the points lifted
//              onto the 4D paraboloid [x4 = x1^2 + x2^2 + x3^2]. So simply
//              a Det5x5 of the points with homogenous coords.
//
//              *** IMPORTANT ***
//               This assumes that the 4 points are ordered correctly according
//               to the left-hand rule, i.e. Orient3d(a,b,c,d) must return TRUE.
//******************************************************************************
class function TGeomTools.InSphere(const a, b, c, d, pt: TPoint): boolean;
var
  re: double;
begin
  inc(numcall_insphere);

  //-- if a-b-c-d are coplanar, then pt is automatically inside the sphere
  if (Orient3D(a, b, c, d) = 0) then
  begin
    Result := true;
  end
  else
  begin
  {$IFDEF FLOAT}

    re := InSphere_value(a, b, c, d, pt);
    if (abs(re) < tolerance) then
      Result := false
    else if (re > 0) then
      Result := true
    else
      Result := false;

  {$ELSEIF True}

    re := InSphereR(a.x, a.y, a.z, b.x, b.y, b.z, c.x, c.y, c.z, d.x, d.y, d.z, pt.x, pt.y, pt.z);
    if (re > 0) then
      Result := true
    else
      Result := false; //-- if directly on the sphere, then return false

  {$IFEND}
  end;
end;


//******************************************************************************
// Description: Determine if a point pt is inside the circumSphere around a
//              tetrahedron. USE OF SYMBOLIC PERTURBATION. The perturbation
//              scheme is as follows. The point with the highest index is
//              perturbed the most in 4D. An arbitrary value is added to its
//              lifted value.
//
// Input:       a, b, c, d : the 4 points forming the tetrahedron
//              pt : the point to test
//
// Output:      TRUE  -> pt is INSIDE the sphere
//              FALSE -> pt is OUTSIDE or ON the sphere
//
// Note:        That's an Orient test made in 4D space with the points lifted
//              onto the 4D paraboloid [x4 = x1^2 + x2^2 + x3^2]. So simply
//              a Det5x5 of the points with homogenous coords.
//
//              *** IMPORTANT ***
//               This assumes that the 4 points are ordered correctly according
//               to the left-hand rule, i.e. Orient3d(a,b,c,d) must return TRUE.
//******************************************************************************
class function  TGeomTools.InSphere_perturb(const a, b, c, d, p: TPoint) : boolean;
//---------------------------------
function compareindex(item1, item2: Pointer) : integer;
var
  a, b: TPoint;
begin
  a := TPoint(item1);
  b := TPoint(item2);
  if (a.m_no < b.m_no) then
    Result := -1
  else if (a.m_no > b.m_no) then
    Result := 1
  else
    Result := 0;
end;
//---------------------------------
const
  pert: double = 0.1;
var
  re: double;
  final: boolean;
  lstIndex: TList;
  a2, b2, c2, d2, p2: double;
  highPt: TPoint;
  continue: boolean;
  pass: integer;
begin
  final := false;
  //-- if a-b-c-d are coplanar, then pt is automatically inside the sphere
  if (Orient3D(a, b, c, d) = 0) then
  begin
    final := true;
  end
  else
  begin
    re := InSphereR(a.x, a.y, a.z, b.x, b.y, b.z, c.x, c.y, c.z, d.x, d.y, d.z, p.x, p.y, p.z);
    if (re > 0) then
      final := true
    else if (re < 0) then
      final := false
    else //-- (re = 0) if result is 0 then perturb the points in (d+1) dimension
    begin
      //-- the point with highest index is perturbed the most upwards, if the
      //-- result is 0 again then the second highest is perturbed, and so on...
      a2 := Sqr(a.x) + Sqr(a.y) + Sqr(a.z);
      b2 := Sqr(b.x) + Sqr(b.y) + Sqr(b.z);
      c2 := Sqr(c.x) + Sqr(c.y) + Sqr(c.z);
      d2 := Sqr(d.x) + Sqr(d.y) + Sqr(d.z);
      p2 := Sqr(p.x) + Sqr(p.y) + Sqr(p.z);

      lstIndex := TList.Create;
      lstIndex.Add(a);
      lstIndex.Add(b);
      lstIndex.Add(c);
      lstIndex.Add(d);
      lstIndex.Add(p);
      lstIndex.Sort(@compareindex);

      continue := true;
      pass := 1;
      while (continue = true) do
      begin
        if (pass = 5) then
        begin
          final := false;
          break;
        end;
        highPt := lstIndex[5 - pass];
        if (highPt = a) then
          a2 := a2 + Math.Power(pert, pass)
        else if (highPt = b) then
          b2 := b2 + Math.Power(pert, pass)
        else if (highPt = c) then
          c2 := c2 + Math.Power(pert, pass)
        else if (highPt = d) then
          d2 := d2 + Math.Power(pert, pass)
        else
          p2 := p2 + Math.Power(pert, pass);

        //-- compute Det5x5 with one point perturbed in 4 dimensions
        re := Det5x5_perturb(a, b, c, d, p, a2, b2, c2, d2, p2);
        if (abs(re) < 1e-12) then //-- floating-point arithmetic, do it again if
          inc(pass)               //-- not sure about result.
        else if (re > 0) then
        begin
          continue := false;
          final := true;
        end
        else if (re < 0) then
        begin
          final := false;
          continue := false;
        end;
      end;
    end;
  end;
  Result := final;
end;


class function  TGeomTools.InSphere_perturb(const te: TTetra; const pt: TPoint) : boolean;
begin
  Result := InSphere_perturb(te.GetP(1), te.GetP(2), te.GetP(3), te.GetP(4), pt);
end;


//******************************************************************************
// Description: Determine if a point pt is above or below the plane defined by
//              a-b-c (anti-clockwise order)
//
// Input:       a,b,c : the points in anti-clockwise order
//              pt : the point to test
//
// Output:      1  -> pt is BELOW of the plane (OK for left-hand rule)
//              0  -> 4 points are coplanar
//             -1  -> pt is ABOVE of the plane (NOT OK for left-hand rule)
//
// Note:        "above and below" means when looking from above;
//              or when using the left-hand rule
//******************************************************************************
class function TGeomTools.Orient3D(const a, b, c, pt: TPoint): integer;
var
  re: double;
begin
  inc(numcall_orient);

{$IFDEF FLOAT}

  //-- floating point arithmetic
  re := Det4x4t(a, b, c, pt);
  if (abs(re) < tolerance) then
    Result := 0
  else if (re > 0) then
    Result := 1
  else //-- (determinant < 0)
    Result := -1;

{$ELSEIF True}

  //-- robust arithmetic
  re := Orient3dR(a.x, a.y, a.z, b.x, b.y, b.z, c.x, c.y, c.z, pt.x, pt.y, pt.z);
  if (re > 0) then
    Result := 1
  else if (re = 0) then
    Result := 0
  else
    Result := -1;

{$IFEND}

end;

//******************************************************************************
// Description: Determine the value of the Orient3d predicate (the value is
//              equal to the 4x4 det of the 4 points )
//
// Input:       a,b,c : the points in anti-clockwise order
//              pt : the point to test
//
// Output:      1  -> pt is BELOW of the plane (OK for left-hand rule)
//              0  -> 4 points are coplanar
//             -1  -> pt is ABOVE of the plane (NOT OK for left-hand rule)
//
// Note:        "above and below" means when looking from above;
//              or when using the left-hand rule
//******************************************************************************
class function TGeomTools.Orient3D_value(const a, b, c, pt: TPoint): double;
begin
  Result := Det4x4t(a, b, c, pt);
end;



//******************************************************************************
// Description: determine if an edge intersects a triangle in 3D
//
// Input:       a, b: the coord of the edge (TPoint)
//              t1, t2, t3: the coord of the triangle (TPoint)
//
// Output:      1  -> (a-b) intersects the triangle
//              0  -> (a-b) is coplanar with 2 points of the triangle
//             -1  -> the edge doesn't intersect the triangle
//              2  -> a is directly on one edge of t1-t2-t3 (flat tetra)
//              3  -> b is directly on one edge of t1-t2-t3 (flat tetra)
//******************************************************************************
class function TGeomTools.IntersectionEdgeTriangle(const a, b, t1, t2,
  t3: TPoint): integer;
var
  orderA: integer;
  orderB: integer;
  test1: integer;
  test2: integer;
  test3: integer;
begin
  orderA := Orient3D(t1, t2, t3, a);
  orderB := Orient3D(t1, t2, t3, b);

  //-- a & b are on the same side of the triangle; or they are both coplanar
  //-- with the triangle
  if (orderA = orderB) then
  begin
    if (orderA = 0) then //-- means that the 5 pts are coplanar
    begin
      raise Exception.Create('5 pts are coplanar'); { TODO : 5 pts are coplanar }
      Result := 1;
    end
    else
    begin
      Result := -1; //-- this case is impossible in my app, but for a general
    end;            //-- case it's possible that both a and b are on the same
  end               //-- side of the triangle

  //-- a & b are on each side of the triangle; or one (only one) can be on the
  //-- plane formed by t1-t2-t3
  else if ( (orderA <> 0) and (orderB <> 0) ) then
  begin
    test1 := Orient3D(b, t2, t3, a);
    test2 := Orient3D(t1, b, t3, a);
    test3 := Orient3D(t1, t2, b, a);
    if ( (test1 = orderB) or (test2 = orderB) or (test3 = orderB) ) then
    begin
      Result := -1;
    end
    else if ( (test1 = 0) or (test2 = 0) or (test3 = 0) ) then
    begin
      Result := 0;
    end
    else
    begin
      Result := 1;
    end;
  end

  //-- a is coplanar with t1-t2-t3 and b is on one side of the triangle
  else if (orderA = 0) then
  begin
    test1 := Orient3D(b, t2, t3, a);
    test2 := Orient3D(t1, b, t3, a);
    test3 := Orient3D(t1, t2, b, a);
    if ( (test1 = orderB) or (test2 = orderB) or (test3 = orderB) ) then
    begin
      Result := -1;
    end
    else if ( (test1 = 0) or (test2 = 0) or (test3 = 0) ) then
    begin   //-- a is directly on one edge of t1-t2-t3
      Result := 2;
    end
    else
    begin
      Result := 1;
    end;
  end
   { TODO : is this case possible? }
  //-- b is coplanar with t1-t2-t3 and a is on one side of the triangle
  else //-- if (orderB = 0) then
  begin
//    raise Exception.Create('Pt b is on the edge!!! -> IntersectionEdgeTriangle');
    test1 := Orient3D(a, t2, t3, b);
    test2 := Orient3D(t1, a, t3, b);
    test3 := Orient3D(t1, t2, a, b);
    if ( (test1 = orderA) or (test2 = orderA) or (test3 = orderA) ) then
    begin
      Result := -1;
    end
    else if ( (test1 = 0) or (test2 = 0) or (test3 = 0) ) then
    begin
      Result := 3;
    end
    else
    begin
      Result := 1;
    end;
  end;
end;


//******************************************************************************
// Description: compute the euclidian distance between 2 points in 3D
//
// Input:       a, b: the coord of the edge (TPoint)
//
// Output:      the distance
//******************************************************************************
class function TGeomTools.Distance3d(const a, b: TPoint): double;
begin
  Result := Sqrt(Sqr(a.x - b.x) + Sqr(a.y - b.y) + Sqr(a.z - b.z));
end;


//******************************************************************************
// Description: compute the normal vector of a triangle, according to a viewpoint.
//              The normal will point in the OPPOSITE direction of this pt
//
// Input:       a, b, c: the 3 vertices of the triangle
//              pt : the view point
//
// Output:      n: the vector pointing in the opposite direction of pt
//******************************************************************************
class procedure TGeomTools.NormalOppositePt(const a, b, c, pt: TPoint; out n: TVector);
var
  v1, v2: TVector;
  temp: TVector;
begin
  //-- creation of 2 vectors on the plane formed by a-b-c
  v1 := TVector.Create(a, c);
  v2 := TVector.Create(a, b);

  //-- we want the normale to point in the direction opposite to pt
  if (Orient3D(a, b, c, pt) <> -1) then
  begin
    temp := v1;
    v1 := v2;
    v2 := temp;
  end;

  //-- calculation of the cross-product
  TVector.CrossProduct(v1, v2, n);
  n.Normalize;
  v1.Free;
  v2.Free;
end;


//******************************************************************************
// Description: compute the normalized normal vector of a triangle, according to
//              a viewpoint.  The normal will point in the DIRECTION of this
//              pt
//
// Input:       a, b, c: the 3 vertices of the triangle
//              pt : the view point
//
// Output:      n: the vector pointing in the direction of pt
//******************************************************************************
class procedure TGeomTools.NormalDirectionPt(const a, b, c, pt: TPoint;
  out n: TVector);
var
  v1, v2: TVector;
  temp: TVector;
begin
  //-- creation of 2 vectors on the plane formed by a-b-c
  v1 := TVector.Create(a, c);
  v2 := TVector.Create(a, b);

  //-- we want the normale to point in the direction opposite to pt
  if (Orient3D(a, b, c, pt) = -1) then
  begin
    temp := v1;
    v1 := v2;
    v2 := temp;
  end;

  //-- calculation of the cross-product
  TVector.CrossProduct(v1, v2, n);
  n.Normalize;
  v1.Free;
  v2.Free;
end;


//******************************************************************************
// Description: Compute the power of a tetrahedron with respect to a point
//
// Input:       a, b, c, d : the 4 vertices defining the tetrahedron
//              v : the point
//
// Output:      value of the pow(v, s) -- s being the circumsphere of tetra
//                negative  means v inside s.
//                positive  means v outside s.
//                zero      means v lying on s.
//
// Remarks:     Power function in 3d is (-InSphere / Orient3d)
//              I put negative to InSphere because it return positive if v is
//              inside the sphere.
//******************************************************************************
class function TGeomTools.Power(const a, b, c, d, v: TPoint): double;
begin
  inc(numcall_orient);
  inc(numcall_insphere);

{$IFDEF FLOAT}

  Result := -InSphere_value(a, b, c, d, v) / Orient3D_value(a, b, c, d);

{$ELSEIF True}

  Result := -InSphereR(a.x, a.y, a.z, b.x, b.y, b.z, c.x, c.y, c.z, d.x, d.y, d.z,
                     v.x, v.y, v.z) /
           Orient3dR(a.x, a.y, a.z, b.x, b.y, b.z, c.x, c.y, c.z, d.x, d.y, d.z);

{$IFEND}
end;


//******************************************************************************
// Description: Calculate the value of the InSphere predicate (the value is the
//              result of a 5x5 matrix with the (d+1)-lifted points to the
//              revolution paraboloid). A translation of -pt to all the points
//              is applied to reduce the computation to a Det4x4.
//
// Input:       a,b,c,d : 4 points defining the sphere
//              pt: the point to test
//
// Output:      value of the 5x5 determinant (a double)
//                positive value -> pt is INSIDE the sphere
//                0 -> pt directly on the sphere
//                negative value -> pt is OUTSIDE the sphere
//******************************************************************************
class function TGeomTools.InSphere_value(const a, b, c, d, pt: TPoint) : double;
var
  ax, ay, az, a2: double;
  bx, by, bz, b2: double;
  cx, cy, cz, c2: double;
  dx, dy, dz, d2: double;
begin
  ax := a.x - pt.x;
  ay := a.y - pt.y;
  az := a.z - pt.z;
  a2 := Sqr(ax) + Sqr(ay) + Sqr(az);
  bx := b.x - pt.x;
  by := b.y - pt.y;
  bz := b.z - pt.z;
  b2 := Sqr(bx) + Sqr(by) + Sqr(bz);
  cx := c.x - pt.x;
  cy := c.y - pt.y;
  cz := c.z - pt.z;
  c2 := Sqr(cx) + Sqr(cy) + Sqr(cz);
  dx := d.x - pt.x;
  dy := d.y - pt.y;
  dz := d.z - pt.z;
  d2 := Sqr(dx) + Sqr(dy) + Sqr(dz);

  ax := ax * Det3x3(by, bz, b2, cy, cz, c2, dy, dz, d2);
  ay := ay * Det3x3(bx, bz, b2, cx, cz, c2, dx, dz, d2);
  az := az * Det3x3(bx, by, b2, cx, cy, c2, dx, dy, d2);
  a2 := a2 * Det3x3(bx, by, bz, cx, cy, cz, dx, dy, dz);

  Result := (ax - ay + az - a2);
end;


//******************************************************************************
// Description: Compute the determinant of a matrix 5x5. This is used to
//              calculate the InSphere_pert(). The InSphere() can usually be
//              reduced to a Det4x4 by a translation but the symbolic perturbation
//              makes that impossible and Det5x5 must be calculated.
//
// Input:       a,b,c,d : 4 points defining the sphere
//              p: the point to test
//              a2..p2: the value (perturbed or not) of the points lifted to a
//              paraboloid in (d+1) dimensions.
//
// Output:      value of the determinant (a double)
//
// Note:        i cannot use robust arithmetic for that since i cannot modify
//              Shewchuk's code. Maybe i could but way too much complicated!
//              And if arithmetic causes problems here just increase the pertub.
//******************************************************************************
class function TGeomTools.Det5x5_perturb(const a, b, c, d, p: TPoint; a2, b2, c2, d2, p2: double) : double;
var
  ax, ay, az, one: double;
begin
  ax := a.x * Det4x4(b.y, b.z, b2, 1,
                     c.y, c.z, c2, 1,
                     d.y, d.z, d2, 1,
                     p.y, p.z, p2, 1);

  ay := a.y * Det4x4(b.x, b.z, b2, 1,
                     c.x, c.z, c2, 1,
                     d.x, d.z, d2, 1,
                     p.x, p.z, p2, 1);

  az := a.z * Det4x4(b.x, b.y, b2, 1,
                     c.x, c.y, c2, 1,
                     d.x, d.y, d2, 1,
                     p.x, p.y, p2, 1);

  a2 := a2 * Det4x4(b.x, b.y, b.z, 1,
                    c.x, c.y, c.z, 1,
                    d.x, d.y, d.z, 1,
                    p.x, p.y, p.z, 1);

  one := Det4x4(b.x, b.y, b.z, b2,
                c.x, c.y, c.z, c2,
                d.x, d.y, d.z, d2,
                p.x, p.y, p.z, p2);

  Result := ax - ay + az - a2 + one;
end;


//******************************************************************************
// Description: Compute the intersection between a line (defined by a starting
//              and an end point) and a sphere.
//
// Input:       p1, p2: start/end points of the line segment
//              te: the 4 points of the tetra defining the sphere
//
// Output:      value of 't' (a double) in the parametric equation
//              [p = p1 + t(p2-p1)]
//
// Note:        Details of the method (solving of a quadratic equation) is
//              available at http://astronomy.swin.edu.au/~pbourke/geometry/sphereline/
//******************************************************************************
class function TGeomTools.IntersectionLineSphere(const p1, p2, pt1, pt2, pt3, pt4: TPoint;
  out t1, t2: double) : boolean;
var
  a, b, c: double;
  p3: TPoint; //-- centre of the sphere
  r2: double; //-- radius of the sphere
  delta: double;
  re: boolean;
begin
  p3 := TPoint.Create;
  self.CircumSphere(pt1, pt2, pt3, pt4, p3);
  r2 := sqr(p3.x - pt1.x) + sqr(p3.y - pt1.y) + sqr(p3.z - pt1.z);

  a := sqr(p2.x - p1.x) + sqr(p2.y - p1.y) + sqr(p2.z - p1.z);
  b := 2 * ( (p2.x - p1.x)*(p1.x - p3.x) + (p2.y - p1.y)*(p1.y - p3.y) + (p2.z - p1.z)*(p1.z - p3.z) );
  c := sqr(p3.x) + sqr(p3.y) + sqr(p3.z) + sqr(p1.x) + sqr(p1.y) + sqr(p1.z)
       - 2 * ( (p3.x * p1.x) + (p3.y * p1.y) + (p3.z * p1.z) ) - r2;

  re := false;
  delta := sqr(b) - (4 * a * c);
  if (abs(delta) < tolerance) then //-- equals to 0
  begin
    ShowMessage('trajectory tangent to a sphere. Do nothing.');
  end
  else if (delta > 0) then //-- 2 solution -> intersections
  begin
    re := true;
    t1 := (-b - sqrt(delta)) / ( 2 * a);
    t2 := (-b + sqrt(delta)) / ( 2 * a);
  end;
  p3.Free;
  Result := re;
end;

class function TGeomTools.IntersectionLineSphere(const p1, p2: TPoint; const te: TTetra; out t1, t2: double) : boolean;
var
  re: boolean;
begin
  re := IntersectionLineSphere(p1, p2, te.GetP(1), te.GetP(2), te.GetP(3), te.GetP(4), t1, t2);
  Result := re;
end;


//******************************************************************************
// Description: Check if the intersection between a line segment and a sphere
//              exists, i.e. by projecting the centre of the sphere onto the
//              line we can determine if this projection is before or after the
//              start point of the segment.
//
// Input:       p1, p2: starting-end points of the line
//              te: the 4 points of the tetra define the sphere
//
// Output:      TRUE  -> projection is after the start point of segment
//              FALSE -> projection is before the start point of segment
//
// Note:        Details of the method (solving of a quadratic equation) is
//              available at http://astronomy.swin.edu.au/~pbourke/geometry/sphereline/
//******************************************************************************
class function TGeomTools.IntersectionLineSphere_projCentre(const p1, p2: TPoint; const te: TTetra) : boolean;
var
  p3: TPoint;
  t: double;
begin
  p3 := TPoint.Create;
  te.GetCircumCentre(p3);
  t := ( (p3.x - p1.x)*(p2.x - p1.x) + (p3.y - p1.y)*(p2.y - p1.y) + (p3.z - p1.z)*(p2.z - p1.z) )
       /
       ( (p2.x - p1.x)*(p2.x - p1.x) + (p2.y - p1.y)*(p2.y - p1.y) + (p2.z - p1.z)*(p2.z - p1.z) );
  p3.Free;
  if (t > 0.0) then
    Result := true
  else
    Result := false;
end;



end.
