unit OpenGLObjIO;

interface

uses
  OpenGLObj, SysUtils, OpenGLTools, OpenGLWritableObject, OpenGLTypes,
  Dialogs;

type

  TGLObjReader = class
    // can throw exception - file not found, incorrect data format, etc.
    class function ReadObjFile(fName: String): TGLObj;
  end;

  TGLObjWriter = class
    // can raise exceptions related to problems with creating/writing to file
    class procedure WriteObjFile(wObject: TGLWritableObject; fName: String);
  end;

implementation

{ TObjReader }

class function TGLObjReader.ReadObjFile(fName: String): TGLObj;
var
  objFile: TextFile;
  vNo, vnNo, vtNo, fNo, gNo: Integer;
  line: String;
  sArray, sArray2: stringArray;
  d1, d2, d3: Double;
  i, i1, i2, i3: Integer;
begin
  sArray := nil;
  sArray2 := nil;
  try
    Assign(objFile, fName);
    Reset(objFile);
    //read file to count items first
    vNo := 0;
    vnNo := 0;
    vtNo := 0;
    fNo := 0;
    gNo := 0;
    while not Eof(objFile) do
    begin
      Readln(objFile, line);
      TStringTool.Truncate(line);
      if (Length(line) > 1) then
      begin
        if line[1] = 'v' then
        begin
          if line[2] = ' ' then
            Inc(vNo)
          else if line[2] = 'n' then
            Inc(vnNo)
          else if line[2] = 't' then
            Inc(vtNo)
        end
        else if line[1] = 'f' then
          Inc(fNo)
        else if line[1] = 'g' then
          Inc(gNo);
      end;
    end;
    //create new TGLObjObject
    Result := TGLObj.Create(vNo, vnNo, vtNo, fNo, gNo);
    //read file again and add all items to TGLObjObject
    Reset(objFile);
    while not Eof(objFile) do
    begin
      Readln(objFile, line);
      sArray := TStringTool.Tokenize(line);
      if sArray <> nil then
      begin
        if sArray[0] = 'v' then
        begin
          if Length(sArray) = 4 then
          begin
            d1 := StrToFloat(sArray[1]);
            d2 := StrToFloat(sArray[2]);
            d3 := StrToFloat(sArray[3]);
            Result.AddVertex(d1, d2, d3);
          end
        end // v
        else if sArray[0] = 'vt' then
        begin
          d1 := 0;
          d2 := 0;
          d3 := -1;
          if Length(sArray) > 2 then
          begin
            d1 := StrToFloat(sArray[1]);
            d2 := StrToFloat(sArray[2]);
          end;
          if Length(sArray) > 3 then
            d3 := StrToFloat(sArray[3]);
          Result.AddVTexture(d1, d2, d3);
        end // vt
        else if sArray[0] = 'vn' then
        begin
          if Length(sArray) = 4 then
          begin
            d1 := StrToFloat(sArray[1]);
            d2 := StrToFloat(sArray[2]);
            d3 := StrToFloat(sArray[3]);
            Result.AddVectorNormal(d1, d2, d3);
          end;
        end // vn
        else if sArray[0] = 'f' then
        begin
          if Length(sArray) > 1 then
            Result.AddFace(Length(sArray)-1);
          for i:=1 to Length(sArray)-1 do
          begin
            i1 := -1;
            i2 := -1;
            i3 := -1;
            sArray2 := TStringTool.Tokenize('/', sArray[i]);
            if Length(sArray2)>0 then
              i1 := StrToInt(sArray2[0]);
            if Length(sArray2)>1 then
              if sArray2[1] <> '' then
                i2 := StrToInt(sArray2[1]);
            if Length(sArray2)>2 then
              if sArray2[2] <> '' then
                i3 := StrToInt(sArray2[2]);
            Result.AddFaceItem(i1, i2, i3);
          end;
        end // f
        else if sArray[0] = 'g' then
        begin
          if Length(sArray) = 2 then
            Result.AddGroup(sArray[1]);
        end // g
      end;
    end;
    CloseFile(objFile);
  except
    on EInOutError do raise;
    on EConvertError do raise;
  end;
end;

{ TGLObjWriter }

class procedure TGLObjWriter.WriteObjFile(wObject: TGLWritableObject; fName: String);
var
  fList: TGLFacesList;
  face: TGLFace;
  objFile: TextFile;
  line: String;
  i: Integer;
  vert: TPoint3;
  vertNo: Integer;
  vect: TVector3;
  vectNo: Integer;
begin
  if wObject <> nil then
  begin
    //opening the output file
    AssignFile(objFile, fName);
    Rewrite(objFile);
    //writing the header
    Writeln(objFile, '# file '+ExtractFileName(fName));
    Writeln(objFile, '# generated on '+FormatDateTime('yyyy-MM-dd HH:mm:ss', Now));
    Writeln(objFile, '# by OpenGL Class Package');
    Writeln(objFile, '#');
    Writeln(objFile, '# LSGI Team');
    Writeln(objFile, '# The Hong Kong Polytechnic University');
    Writeln(objFile, '# Department of Land-Surveying & Geo-Informatics');
    Writeln(objFile, '#');
    //creating the list of faces
    fList := wObject.GetFacesList;
    //scan list three times, first to write all vertices, second time
    //to write all vector normals, third  time to write all faces
    face := fList.GetFirstFace;
    Writeln(objFile, '# vertices');
    while face <> nil do
    begin
      for i:=0 to face.vCount-1 do
      begin
        vert := face.vertices[i];
        line := 'v '+FloatToStr(vert.x)+' '+FloatToStr(vert.y)+' '+FloatToStr(vert.z);
        Writeln(objFile, line);
      end;
      face := fList.GetNextFace;
    end;
    //second scan of the list
    face := fList.GetFirstFace;
    Writeln(objFile, '# normal vectors');
    while face <> nil do
    begin
      for i:=0 to face.vCount-1 do
        if face.vNormals then
        begin
          vect := face.vectors[i];
          line := 'vn '+FloatToStr(vect.x)+' '+FloatToStr(vect.y)+' '+FloatToStr(vect.z);
          Writeln(objFile, line);
        end;
      face := fList.GetNextFace;
    end;
    //third scan of the list
    face := fList.GetFirstFace;
    vertNo := 1;
    vectNo := 1;
    Writeln(objFile, '# faces');
    while face <> nil do
    begin
      line := 'f ';
      for i:=0 to face.vCount-1 do
      begin
        line := line+IntToStr(vertNo);
        Inc(vertNo);
        if face.vNormals then
        begin
          line := line+'//'+IntToStr(vectNo);
          Inc(vectNo);
        end;
        line := line+' ';
      end;
      Writeln(objFile, line);
      face := fList.GetNextFace;
    end;
    //close the output file
    CloseFile(objFile);
  end; //if wObject <> nil
end;

end.
