unit OpenGLScene;

interface

uses
  OpenGL, Classes, Windows, SysUtils, Controls, Forms,
  OpenGLCamera, OpenGLObject, OpenGLLight,
  OpenGLManipulator, OpenGLTypes;

type

  TGLScene = class
  private
    //fields to initialize openGL
    OpenGLReady: Boolean;
    GLContext: HGLRC;
    glDC: HDC;
    //references to form, and drawing area
    handle: HWND; //drawing area handle (form handle, paintbox handle, etc.)
    drawArea: TControl; //(form, paintbox, panel, etc.)
    form: TForm; //form which contains drawing area
    //components of the scene
    camera: TGLCamera;
    manipulator: TGLManipulator;
    terrain: TGLObject;
    objects: TList;
    lights: TList;
    //values of left, right, top, and bottom gaps, useful when drawing directly
    //on the form, when using also panels with buttons, etc.
    //panels allignment should be set to all top/right/left/bottom, and gaps values
    //set to panels width/height
    leftGap, rightGap, topGap, bottomGap: Integer;
  public
    constructor Create(handle: HWND; drawArea: TControl; form: TForm);
    procedure SetTerrain(terrain: TGLObject);
    procedure SetManipulator(manipulator: TGLManipulator);
    destructor Destroy; override;
    procedure AddLight(light: TGLLight);
    procedure AddObject(obj: TGLObject);
    function getLightsNumber: Integer;
    function getLight(lightNum: Integer): pTGLLight;
    procedure TurnLight(lightNum: Integer; switch: Boolean);
    procedure TurnAllLights(switch: Boolean);
//    procedure SetProperties; virtual;
    procedure Initialize; //initialization of OpenGL
    procedure DrawObjects;
    function getDrawArea: TControl;
    function getManipulator: TGLManipulator;
    function getCamera: TGLCamera;
    procedure Reset;
    procedure SetGaps(left, top, right, bottom: Integer);
    //keyboard and mouse events handling, return value inicates whether drawing area should be repainted
    function OnKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState): TGLEventResult;
    function OnKeyPress(Sender: TObject; var Key: Char): TGLEventResult;
    function OnKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState): TGLEventResult;
    function OnMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer): TGLEventResult;
    function OnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer): TGLEventResult;
    function OnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer): TGLEventResult;
    function OnMouseWheel(Sender: TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean): TGLEventResult;
    function OnMouseWheelDown(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean): TGLEventResult;
    function OnMouseWheelUp(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean): TGLEventResult;
    function OnFormResize(Sender: TObject): TGLEventResult;
    function getHDC : HDC;
  end;

implementation

//uses

{ TOpenGLScene }

procedure TGLScene.AddLight(light: TGLLight);
begin
  lights.Add(light);
end;

constructor TGLScene.Create(handle: HWND; drawArea: TControl; form: TForm);
begin
  self.handle := handle;
  self.drawArea := drawArea;
  self.form := form;
  lights := TList.Create;
  objects := TList.Create;
  camera := TGLCamera.Create;
end;

procedure TGLScene.TurnAllLights(switch: Boolean);
var
  light: TGLLight;
  i: Integer;
begin
  for i:=0 to Lights.Count-1 do
  begin
    if switch then
    begin
      light := Lights.Items[i];
      glLightfv(GL_LIGHT0+i, GL_AMBIENT, @(light.ambient));
      glLightfv(GL_LIGHT0+i, GL_DIFFUSE, @(light.diffuse));
      glLightfv(GL_LIGHT0+i, GL_SPECULAR, @(light.specular));
      glLightfv(GL_LIGHT0+i, GL_POSITION, @(light.position));
      glEnable(GL_LIGHT0+i);
    end
    else
      glDisable(GL_LIGHT0+i);
  end;
end;

procedure TGLScene.TurnLight(lightNum: Integer; switch: Boolean);
var
  light: TGLLight;
begin
  if not (lightNum > Lights.Count) then
    if switch then
    begin
      light := Lights.Items[lightNum];
      glLightfv(GL_LIGHT0+lightNum, GL_AMBIENT, @(light.ambient));
      glLightfv(GL_LIGHT0+lightNum, GL_DIFFUSE, @(light.diffuse));
      glLightfv(GL_LIGHT0+lightNum, GL_SPECULAR, @(light.specular));
      glLightfv(GL_LIGHT0+lightNum, GL_POSITION, @(light.position));
      glEnable(GL_LIGHT0+lightNum);
    end
    else
      glDisable(GL_LIGHT0+lightNum);
end;

function TGLScene.getLight(lightNum: Integer): pTGLLight;
begin
  Result := Lights.Items[lightNum];
end;

function TGLScene.getLightsNumber: Integer;
begin
  Result := Lights.Count;
end;

procedure TGLScene.Initialize;
var
  pfd: TPixelFormatDescriptor;
  FormatIndex: integer;
begin
  fillchar(pfd, SizeOf(pfd), 0);
  //setting appropriate values to pixel format descriptor attributes
  with pfd do
  begin
    nSize       := SizeOf(pfd);
    nVersion    := 1; //current version of the descriptor
    dwFlags     := PFD_DRAW_TO_WINDOW or PFD_SUPPORT_OPENGL or PFD_DOUBLEBUFFER;
    iPixelType  := PFD_TYPE_RGBA;
    cColorBits  := 24;  //24 bits colors
    cDepthBits  := 32;  //depth of z-buffer
    iLayerType  := PFD_MAIN_PLANE;
  end;  //with
  // getting the handle to component using to draw on
  glDC := getDC(handle);
  // setting the pixel format for this component
  FormatIndex := ChoosePixelFormat(glDC, @pfd);
  //checking if operation was successful
  if FormatIndex = 0 then
    raise Exception.Create('Failure in ChoosePixelFormat '+IntToStr(GetLastError));
  // setting choosen pixel format and checking if operation succeded
  if not SetPixelFormat(glDC, FormatIndex, @pfd) then
    raise Exception.Create('Failure in SetPixelFormat '+IntToStr(GetLastError));
  // creating context
  GLContext := wglCreateContext(glDC);
  // checking if creation of context was ok
  if GLContext = 0 then
    raise Exception.Create('Failure in wglCreateContext '+IntToStr(GetLastError));
  // setting created context as current and checking
  if not wglMakeCurrent(glDC, GLContext) then
    raise Exception.Create('Failure in wglMakeCurrent '+IntToStr(GetLastError));
  // if no problem occured openGL is ready
  openGLReady := true;
end;

{
procedure TGLScene.SetProperties;
var
  ambientLight: TLightParameters;
begin
  glEnable(GL_DEPTH_TEST);
  // enabling hiding the back surfaces
  //glEnable(GL_CULL_FACE);
  //glCullFace(GL_BACK);
  // enabling smooth shade model
//  glShadeModel(GL_SMOOTH);
  // enable lighting
  glEnable(GL_LIGHTING);
  // enable material properties
  //glEnable(GL_COLOR_MATERIAL);
  //set additional ambient light in the scene

  ambientLight.r := 0.5;
  ambientLight.g := 0.5;
  ambientLight.b := 0.5;
  ambientLight.a := 1.0;
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @ambientLight);
  //set camera initial properties
  camera.SetPosition(0, 0, 3, 0, 0, 0, 0, 1, 0);
  camera.SetShape(60, 1, 0.1, 25);

end;
}

destructor TGLScene.Destroy;
var
  i: Integer;
  light: TGLLight;
begin
  wglMakeCurrent(glDC, 0);
  ReleaseDC(handle, glDC);
  wglDeleteContext(GLContext);
  camera.Destroy;
  for i:=0 to lights.Count-1 do
  begin
    light := lights.Items[i];
    light.Free;
  end;
  lights.Destroy;
  objects.Destroy;
  manipulator.Destroy;
  inherited;
end;

procedure TGLScene.DrawObjects;
var
  i, errorCode: Integer;
  obj: TGLObject;
begin
  if not OpenGLReady then
    raise Exception.Create('OpenGL has not been initialized');
  //reseting the modelview matrix, and setting lookAt, by calling camera's routine
  camera.SetModelViewMatrix;
  // buffers clearing
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  //translations, etc. to be done by manipulator
  manipulator.Manipulate;
  //terrain drawing
  terrain.Draw;
  //all enabled objects drawing
  for i:=0 to objects.Count-1 do
  begin
    obj := objects.Items[i];
    if obj.enabled then
      obj.Draw;
  end;
  //other drawings
  //...
  // swapping of buffers
  SwapBuffers(glDC);
  //  Error handling
  errorCode := glGetError;
  if errorCode <> GL_NO_ERROR then
    raise Exception.Create('Error occured during drawing: '+gluErrorString(errorCode));
end;

function TGLScene.OnKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState): TGLEventResult;
begin
  if manipulator = nil then
    Result := glerNotHandled
  else
    Result := manipulator.OnKeyDown(Sender, Key, Shift);
end;

function TGLScene.OnKeyPress(Sender: TObject; var Key: Char): TGLEventResult;
begin
  if manipulator = nil then
    Result := glerNotHandled
  else
    Result := manipulator.OnKeyPress(Sender, Key);
end;

function TGLScene.OnKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState): TGLEventResult;
begin
  if manipulator = nil then
    Result := glerNotHandled
  else
    Result := manipulator.OnKeyUp(Sender, Key, Shift);
end;

function TGLScene.OnMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer): TGLEventResult;
begin
  if manipulator = nil then
    Result := glerNotHandled
  else
    Result := manipulator.OnMouseDown(Sender, Button, Shift, X, Y);
end;

function TGLScene.OnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer): TGLEventResult;
begin
  if manipulator = nil then
    Result := glerNotHandled
  else
    Result := manipulator.OnMouseMove(Sender, Shift, X, Y);
end;

function TGLScene.OnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer): TGLEventResult;
begin
  if manipulator = nil then
    Result := glerNotHandled
  else
    Result := manipulator.OnMouseUp(Sender, Button, Shift, X, Y);
end;

function TGLScene.OnMouseWheel(Sender: TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean): TGLEventResult;
begin
  if manipulator = nil then
    Result := glerNotHandled
  else
    Result := manipulator.OnMouseWheel(Sender, Shift, WheelDelta, MousePos, Handled);
end;

function TGLScene.OnMouseWheelDown(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean): TGLEventResult;
begin
  if manipulator = nil then
    Result := glerNotHandled
  else
    Result := manipulator.OnMouseWheelDown(Sender, Shift, MousePos, Handled);
end;

function TGLScene.OnMouseWheelUp(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean): TGLEventResult;
begin
  if manipulator = nil then
    Result := glerNotHandled
  else
    Result := manipulator.OnMouseWheelUp(Sender, Shift, MousePos, Handled);
end;

procedure TGLScene.SetManipulator(manipulator: TGLManipulator);
begin
  self.manipulator := manipulator;
  manipulator.SetCamera(camera);
  manipulator.SetGaps(leftGap, topGap, rightGap, bottomGap);
  manipulator.Initialize;
end;

function TGLScene.getDrawArea: TControl;
begin
  Result := drawArea;
end;

function TGLScene.OnFormResize(Sender: TObject): TGLEventResult;
begin
  if ((drawArea.ClientWidth-leftGap-rightGap) >= 0) and ((drawArea.ClientHeight-topGap-bottomGap) >= 0 ) then
    glViewport(leftGap, bottomGap, drawArea.ClientWidth-leftGap-rightGap, drawArea.ClientHeight-topGap-bottomGap)
  else
    glViewport(leftGap, bottomGap, drawArea.ClientWidth, drawArea.ClientHeight);
  Result := glerRepaintNeeded;
end;

procedure TGLScene.SetTerrain(terrain: TGLObject);
begin
  self.terrain := terrain;
end;

procedure TGLScene.Reset;
begin
  manipulator.Reset;
end;

procedure TGLScene.AddObject(obj: TGLObject);
begin
  objects.Add(obj);
end;

function TGLScene.getManipulator: TGLManipulator;
begin
  Result := self.manipulator;
end;

function TGLScene.getCamera: TGLCamera;
begin
  Result := self.camera;
end;

procedure TGLScene.SetGaps(left, top, right, bottom: Integer);
begin
  self.leftGap := left;
  self.topGap := top;
  self.rightGap := right;
  self.bottomGap := bottom;
end;

function TGLScene.getHDC: HDC;
begin
  Result := glDC;
end;

end.
