unit OpenGLWritableObject;

interface

uses
  OpenGLObject, OpenGLTypes, Classes;

type

  TGLFace = class
    vNormals: Boolean;
    vertices: array of TPoint3;
    vectors: array of TVector3;
    vCount: Integer;
    constructor Create(vNormals: Boolean); overload;
    constructor Create(vertNo: Integer; vNormals: Boolean); overload;
    procedure AddVertex(x, y, z: Double);
    procedure AddVertexWithNormal(x, y, z, a, b, c: Double);
    destructor Destroy; override;
  end;

  TGLFacesList = class
  private
    list: TList;
    currentIndex: Integer;
  public
    constructor Create; overload;
    constructor Create(capacity: Integer); overload;
    procedure AddFace(face: TGLFace);
    function GetFirstFace: TGLFace;
    function GetNextFace: TGLFace; //returns TGLFace, or nil if after last face
    function GetFace(index: Integer): TGLFace; //returns TGLFace, or nil
    function GetCount: Integer;
    destructor Destroy; override;
  end;

  TGLWritableObject = class(TGLObject)
  public
    function GetFacesList: TGLFacesList; virtual; abstract;
  end;

implementation

{ TGLFace }

procedure TGLFace.AddVertex(x, y, z: Double);
var
  vert: TPoint3;
begin
  vert := TPoint3.Create(x, y, z);
  if High(vertices) = vCount-1 then
    SetLength(vertices, vCount+3);
  vertices[vCount] := vert;
  Inc(vCount);
end;

procedure TGLFace.AddVertexWithNormal(x, y, z, a, b, c: Double);
var
  vert: TPoint3;
  vect: TVector3;
begin
  vert := TPoint3.Create(x, y, z);
  vect := TVector3.Create(a, b, c);
  if High(vertices) = vCount-1 then
    SetLength(vertices, vCount+3);
  if High(vectors) = vCount-1 then
    SetLength(vectors, vCount+3);
  vertices[vCount] := vert;
  vectors[vCount] := vect;
  Inc(vCount);
end;

constructor TGLFace.Create(vNormals: Boolean);
begin
  vCount := 0;
  SetLength(vertices, 3);
  if vNormals then
  begin
    self.vNormals := true;
    SetLength(vectors, 3);
  end;
end;

constructor TGLFace.Create(vertNo: Integer; vNormals: Boolean);
begin
  vCount := 0;
  SetLength(vertices, vertNo);
  begin
    self.vNormals := true;
    SetLength(vectors, vertNo);
  end;
end;

destructor TGLFace.Destroy;
var
  i: Integer;
begin
  for i:=0 to vCount-1 do
  begin
    vertices[i].Destroy;
    if vNormals then
      vectors[i].Destroy;
  end;
  inherited;
end;

{ TGLFacesList }

constructor TGLFacesList.Create;
begin
  list := TList.Create;
  list.Capacity := 10;
end;

constructor TGLFacesList.Create(capacity: Integer);
begin
  list := TList.Create;
  list.Capacity := capacity;
end;

procedure TGLFacesList.AddFace(face: TGLFace);
begin
  if face <> nil then
    list.Add(face);
end;

function TGLFacesList.GetFirstFace: TGLFace;
begin
  if list.Count > 0 then
  begin
    Result := list.Items[0];
    currentIndex := 1;
  end
  else
    Result := nil;
end;

function TGLFacesList.GetNextFace: TGLFace;
begin
  if list.Count > currentIndex then
  begin
    Result := list.Items[currentIndex];
    Inc(currentIndex);
  end
  else
    Result := nil;
end;

function TGLFacesList.GetFace(index: Integer): TGLFace;
begin
  if list.Count > index then
    Result := list.Items[index]
  else
    Result := nil;
end;

function TGLFacesList.GetCount: Integer;
begin
  Result := list.Count;
end;

destructor TGLFacesList.Destroy;
var
  i: Integer;
  face: TGLFace;
begin
  for i:=0 to list.Count do
  begin
    face := list.Items[i];
    face.Destroy;
  end;
  list.Free;
end;

end.
