//*****************************************************************************
// File:        Tetra.pas
//
// Classe:      TTetra
//
// Description: TTetra represents one tetra, with 4 pointers to the 4 vertices
//              and 4 pointers to the adjacent tetra.
//
// Note:
//
//*****************************************************************************
// Last version: 2003-09-13
//*****************************************************************************

unit Tetra;

{$IFDEF DEBUG}
  {$ASSERTIONS ON}
{$ELSEIF True}
  {$ASSERTIONS OFF}
{$IFEND}

interface

uses
  Point;

type
  TTetra = class
  private
    m_points:   array[1..4] of TPoint;
    m_tetra:    array[1..4] of TTetra;
    m_centre:   TPoint;
    m_cur:      integer; //-- current vertex/face in the tetra
    function    Rot(i, j: integer) : integer;
    procedure   TestInvariants;
  public
    m_no:       integer;
    constructor Create(v1, v2, v3, v4: TPoint);
    destructor  Destroy; override;    
    function    GetP(i: integer) : TPoint;
    function    GetProt(r: integer) : TPoint;
    procedure   SetP(i: integer; pt: TPoint);
    procedure   SetProt(r: integer; pt: TPoint);
    function    GetT(i: integer) : TTetra;
    function    GetTrot(r: integer) : TTetra;
    procedure   SetT(i: integer; te: TTetra);
    procedure   SetTrot(r: integer; te: TTetra);

    function    SetIndex(i: integer) : TTetra; overload;
    function    SetIndex(const pt: TPoint) : TTetra; overload;
    function    SetIndex(const te: TTetra) : TTetra; overload;
    function    RotIndex(r: integer) : TTetra;
    function    GetIndex : integer; overload;
    function    GetIndex(const pt: TPoint) : integer; overload;
    function    GetIndex(const te: TTetra) : integer; overload;

    procedure   ResetCentre;

    function    HasPoint(const pt: TPoint) : boolean;
    function    HasNeighbour(const te: TTetra) : boolean;
    function    HasEdge(const org, dest: TPoint) : boolean;
    function    HasFace(const pt1, pt2, pt3: TPoint): boolean;

    function    IsOrient3D : boolean;
    function    IsFlat : boolean;                    
    procedure   GetCircumCentre(out centre: TPoint);
    function    GetVolume : double;
  end;


implementation

uses
  GeomTools;


//-- static variable (to debug)
var
  no: integer = 0;

{ TTetra }


//******************************************************************************
// Description: Constructor of a tetra. The vertices are needed, and the
//              neighbours are set to NIL.
//
// Input:       the 4 vertices, correctly oriented according to Orient3d()
//
// Output:      -
//******************************************************************************
constructor TTetra.Create(v1, v2, v3, v4: TPoint);
var
  i: integer;
begin
  m_points[1] := v1;
  m_points[2] := v2;
  m_points[3] := v3;
  m_points[4] := v4;
  for i := 1 to 4 do
  begin
    m_tetra[i] := nil;
  end;
  m_no := no;
  inc(no);
  m_cur := 1;
  m_centre := TPoint.Create(-999, -999, -999);
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
end;


//******************************************************************************
// Description: Get one vertex in the tetra.
//
// Input:       the index of the vertex in the array of the tetra [1..4]
//
// Output:      the vertex (a TPoint)
//******************************************************************************
function TTetra.GetP(i: integer) : TPoint;
begin
  Assert( (i >= 1) and (i <= 4) );
  Result := m_points[i];
end;


//******************************************************************************
// Description: Get one vertex in the tetra.
//
// Input:       r: the number of rotation (0, 1, 2 or 3) starting from m_cur
//
// Output:      the vertex (a TPoint)
//******************************************************************************
function TTetra.GetProt(r: integer): TPoint;
begin
  Assert( (r >= 0) and (r <= 3) );
  Result := m_points[Rot(r, m_cur)];
end;


//******************************************************************************
// Description: Set one vertex of the tetra to a new value.
//
// Input:       i : the index of the vertex in the array of the tetra [1..4]
//              pt: the point
//
// Output:      -
//******************************************************************************
procedure TTetra.SetP(i: integer; pt: TPoint);
begin
  Assert( (i >= 1) and (i <= 4) );
  m_points[i] := pt;
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
end;


//******************************************************************************
// Description: Set one vertex of the tetra to a new value.
//
// Input:       r: the number of rotation (0, 1, 2 or 3) starting from m_cur
//              pt: the point
//
// Output:      -
//******************************************************************************
procedure TTetra.SetProt(r: integer; pt: TPoint);
begin
  Assert( (r >= 0) and (r <= 3) );
  m_points[Rot(r, m_cur)] := pt;
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
end;


function TTetra.GetT(i: integer) : TTetra;
begin
  Assert( (i >= 1) and (i <= 4) );
  Result := m_tetra[i];
end;


function TTetra.GetTrot(r: integer): TTetra;
begin
  Assert( (r >= 0) and (r <= 3) );
  Result := m_tetra[Rot(r, m_cur)];
end;


procedure TTetra.SetT(i: integer; te: TTetra);
begin
  Assert( (i >= 1) and (i <= 4) );
  m_tetra[i] := te;
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
end;


procedure TTetra.SetTrot(r: integer; te: TTetra);
begin
  Assert( (r >= 0) and (r <= 3) );
  m_tetra[Rot(r, m_cur)] := te;
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
end;


//******************************************************************************
// Description: Set the current vertex-neighbour to a certain position in the
//              the array m_points[]
//
// Input:       i : integer between [1..4]
//
// Output:      the tetra with an updated m_cur
//******************************************************************************
function TTetra.SetIndex(i: integer) : TTetra;
begin
  Assert( (i >= 1) and (i <= 4) );
  m_cur := i;
  Result := self;
end;


//******************************************************************************
// Description: Set the current vertex-neighbour to a certain TPoint
//
// Input:       pt : the point
//
// Output:      the tetra with an updated m_cur if pt was one of the vertex of
//              the tetra; if not the old m_cur is kept
//******************************************************************************
function TTetra.SetIndex(const pt: TPoint) : TTetra;
var
  index: integer;
begin
  index := self.GetIndex(pt);
  Assert(index <> 0);
  m_cur := index;
  Result := self;
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
end;


//******************************************************************************
// Description: Set the current vertex-neighbour to the opposite of a TTetra
//
// Input:       te : the tetra to find in the neighbouring list
//
// Output:      the tetra with an updated m_cur if te is a neighbours of self
//******************************************************************************
function TTetra.SetIndex(const te: TTetra) : TTetra;
var
  index: integer;
begin
  index := GetIndex(te);
  Assert(index <> 0);
  if (index <> 0) then
  begin
    m_cur := index;
  end;
  Result := self;
  {$IFDEF DEBUG}
  TestInvariants;
  {$ENDIF}
end;


//******************************************************************************
// Description: Determine is an edge is present in the tetra
//
// Input:       the origin and destination of the edge
//
// Output:      TRUE  -> edge is present
//              FALSE -> edge is NOT there
//
// Remarks:     the directtion of the edge is not important. This is just
//              testing whether the 2 points are in the tetra.
//******************************************************************************
function TTetra.HasEdge(const org, dest: TPoint): boolean;
begin
  if ( (self.HasPoint(org) = true) and (self.HasPoint(dest) = true) ) then
  begin
    Result := true;
  end
  else
  begin
    Result := false;
  end;
end;


//******************************************************************************
// Description: Determine if the 4 vertices of a tetra are "correctly ordered"
//              (according to the Orient3D test in the GeomTools unit)
//
// Input:       -
//
// Output:      TRUE  -> tetra is correctly ordered or the 4 points are coplanar
//              FALSE -> tetra is not correctly ordered
//******************************************************************************
function TTetra.IsOrient3D: boolean;
var
  orient: integer;
begin
  orient := TGeomTools.Orient3D(m_points[1], m_points[2], m_points[3], m_points[4]);
  if (orient = -1) then
    Result := false
  else
    Result := true;
end;


procedure TTetra.GetCircumCentre(out centre: TPoint);
begin
//  if (m_centre.x = -999) then
//  begin
    TGeomTools.CircumSphere(m_points[1], m_points[2], m_points[3], m_points[4], centre);
{    m_centre.CopyValue(centre);
  end
  else
  begin
    centre.x := m_centre.x;
    centre.y := m_centre.y;
    centre.z := m_centre.z;
  end;
}end;


//******************************************************************************
// Description: Determine if a point is one of the vertices of a tetra.
//
// Input:       pt: the point to check
//
// Output:      TRUE  -> point is there
//              FALSE -> point is NOT there
//******************************************************************************
function TTetra.HasPoint(const pt: TPoint): boolean;
begin
  if (GetIndex(pt) <> 0) then
    Result := true
  else
    Result := false;
end;


//******************************************************************************
// Description: Determine if a tetra is one of the neighbours of a tetra.
//
// Input:       te: the tetra to check
//
// Output:      TRUE  -> te is a neighbour
//              FALSE -> te is NOT a neighbour
//******************************************************************************
function TTetra.HasNeighbour(const te: TTetra): boolean;
var
  i: integer;
  found: boolean;
begin
  found := false;
  for i := 1 to 4 do
  begin
    if (m_tetra[i] = te) then
    begin
      found := true;
      break;
    end;
  end;
  Result := found;
end;


//******************************************************************************
// Description: to "rotate" in a tetrahedron (from one vertex to another; or
//              from one edge to another)
//
// Input:       i : the number of rotations (0 (same place), 1, 2 or 3)
//              j : the starting position (an integer from 1 to 4)
//
// Output:      the new position (an integer from 1 to 4)
//******************************************************************************
function TTetra.Rot(i, j: integer): integer;
const
  a: array[0..3, 1..4] of integer = ((1, 2, 3, 4), (2, 3, 4, 1),
                                     (3, 4, 1, 2), (4, 1, 2, 3));
begin
  Assert( (i >= 0) and (i <= 3) and (j >= 1) and (j <= 4) );
  Result := a[i, j];
end;


//******************************************************************************
// Description: Get the index of a point in the tetra
//
// Input:       pt: the point whose index is asked
//
// Output:      the index [1..4] if pt is a vertex of the tetra
//              0 if pt is NOT a vertex
//******************************************************************************
function TTetra.GetIndex(const pt: TPoint) : integer;
var
  i: integer;
  r: integer;
begin
  r := 0;
  for i := 1 to 4 do
  begin
    if (m_points[i] = pt) then
    begin
      r := i;
      break;
    end;
  end;
  Result := r;
end;


//******************************************************************************
// Description: Get the index of a tetra (neighbour) in the tetra
//
// Input:       te: the tetra whose index is asked
//
// Output:      the index [1..4] if te is a neighbour of the tetra
//              0 if te is NOT a neighbour
//******************************************************************************
function TTetra.GetIndex(const te: TTetra) : integer;
var
  i: integer;
  r: integer;
begin
  r := 0;
  for i := 1 to 4 do
  begin
    if (m_tetra[i] = te) then
    begin
      r := i;
      break;
    end;
  end;
  Result := r;
end;


//******************************************************************************
// Description: Get the current index of the tetra
//
// Input:       -
//
// Output:      the current index (the value of m_cur which is [1..4]
//******************************************************************************
function TTetra.GetIndex: integer;
begin
  Result := m_cur;
end;


//******************************************************************************
// Description: Determine if a tetra is flat - if the 4 vertices are coplanar
//
// Input:       -
//
// Output:      TRUE  -> te is flat
//              FALSE -> te is NOT flat
//******************************************************************************
function TTetra.IsFlat: boolean;
begin
  if (TGeomTools.Orient3D(m_points[1], m_points[2], m_points[3],
                          m_points[4]) = 0) then
  begin
    Result := true;
  end
  else
  begin
    Result := false;
  end;
end;


//******************************************************************************
// Description: Rotate the m_cur of the tetra by rotating by [0..3] rotations.
//
// Input:       r: number of rotations
//
// Output:      the tetra with a m_cur updated.
//******************************************************************************
function TTetra.RotIndex(r: integer): TTetra;
begin
  Assert( (r > 0) and (r < 3) );
  m_cur := Rot(r, m_cur);
  Result := self;
end;


procedure TTetra.TestInvariants;
var
  i: integer;
begin
  Assert(self.IsOrient3D = true);
  for i := 1 to 4 do
  begin
    Assert(m_points[i] <> nil);
  end;
  Assert( (m_cur >= 1) and (m_cur <= 4) );
end;


function TTetra.GetVolume: double;
var
  re: double;
begin
  re := TGeomTools.Det4x4t(m_points[1], m_points[2], m_points[3], m_points[4]);
  Result := re / 6;
end;


destructor TTetra.Destroy;
begin
  m_centre.Free;
  inherited Destroy;
end;


procedure TTetra.ResetCentre;
begin
  m_centre.SetCoord(-999, -999, -999);
end;


function TTetra.HasFace(const pt1, pt2, pt3: TPoint): boolean;
begin
  if ( (self.HasPoint(pt1) = true) and
       (self.HasPoint(pt2) = true) and
       (self.HasPoint(pt3) = true) ) then
  begin
    Result := true;
  end
  else
  begin
    Result := false;
  end;
end;



end.
