//*****************************************************************************
// File:        Vector.pas
//
// Classe:      TVector
//
// Description: TVector represents a vector in 3d. The usual operations are
//              possible.
//
// Note:
//
//*****************************************************************************
// Last version: 25-06-2002
//*****************************************************************************

unit Vector;

interface

uses
  Point,
  SysUtils;

type
  TVector = class
  private
    m_coord:        TPoint;
    procedure       SetCoord(const coord: TPoint);
  public
    constructor     Create; overload;
    constructor     Create(dest: TPoint); overload;
    constructor     Create(destX, destY, destZ: double); overload;
    constructor     Create(orig, dest: TPoint); overload;
    destructor      Destroy; override;
    procedure       Add(const a: TVector);
    procedure       Subtract(const a: TVector);
    procedure       Multiply(a: double);
    function        GetLength : double;
    procedure       Normalize;
    class procedure CrossProduct(const a, b: TVector; out r: TVector);
    class function  DotProduct(const a, b: TVector) : double;
  published
    property coord: TPoint read m_coord write SetCoord;
  end;

implementation

{ TVector }


constructor TVector.Create;
begin
  m_coord := TPoint.Create(0, 0, 0);
end;


constructor TVector.Create(dest: TPoint);
begin
  m_coord := TPoint.Create(dest.x, dest.y, dest.z);
end;


constructor TVector.Create(destX, destY, destZ: double);
begin
  m_coord := TPoint.Create(destX, destY, destZ);
end;


//******************************************************************************
// Description: Constructor of TVector
//
// Input:       orig and dest of the vector
//
// Output:      -
//
// Remarks:     only the length and the direction is preserved, so the orig of
//              every edge stored is (0, 0, 0)
//******************************************************************************
constructor TVector.Create(orig, dest: TPoint);
begin
  m_coord := TPoint.Create(dest.x - orig.x, dest.y - orig.y, dest.z - orig.z);
end;


destructor TVector.Destroy;
begin
  m_coord.Free;
  inherited Destroy;
end;


procedure TVector.Add(const a: TVector);
begin
  m_coord.x := m_coord.x + a.m_coord.x;
  m_coord.y := m_coord.y + a.m_coord.y;
  m_coord.z := m_coord.z + a.m_coord.z;
end;


class procedure TVector.CrossProduct(const a, b: TVector; out r: TVector);
begin
  r.m_coord.x := (a.m_coord.y * b.m_coord.z) - (a.m_coord.z * b.m_coord.y);
  r.m_coord.y := -((a.m_coord.x * b.m_coord.z) - (a.m_coord.z * b.m_coord.x));
  r.m_coord.z := (a.m_coord.x * b.m_coord.y) - (a.m_coord.y * b.m_coord.x);
end;


procedure TVector.Multiply(a: double);
begin
  m_coord.x := m_coord.x * a;
  m_coord.y := m_coord.y * a;
  m_coord.z := m_coord.z * a;
end;


procedure TVector.Subtract(const a: TVector);
begin
  m_coord.x := m_coord.x - a.m_coord.x;
  m_coord.y := m_coord.y - a.m_coord.y;
  m_coord.z := m_coord.z - a.m_coord.z;
end;


function TVector.GetLength: double;
begin
  Result := Sqrt((m_coord.x * m_coord.x) +
                 (m_coord.y * m_coord.y) +
                 (m_coord.z * m_coord.z));
end;


procedure TVector.SetCoord(const coord: TPoint);
begin
  m_coord.x := coord.x;
  m_coord.y := coord.y;
  m_coord.z := coord.z;
end;


//******************************************************************************
// Description: Normalize the vector
//
// Input:       -
//
// Output:      - (the vector is normalized)
//
// Remarks:     only the length and the direction is preserved, so the orig of
//              every edge stored is (0, 0, 0)
//******************************************************************************
procedure TVector.Normalize;
const
  tolerance: double = 1e-12;
var
  length: double;
begin
  length := self.GetLength;
  //-- raise an EZeroDivide if the lenght of the vector is 0 (or near 0)
  if (abs(length) < tolerance) then
  begin
    raise EZeroDivide.Create('Can not normalise vector, EZeroDivide.');
  end;
  m_coord.x := m_coord.x / length;
  m_coord.y := m_coord.y / length;
  m_coord.z := m_coord.z / length;
end;


class function TVector.DotProduct(const a, b: TVector): double;
begin
  Result := ( (a.m_coord.x * b.m_coord.x) +
              (a.m_coord.y * b.m_coord.y) +
              (a.m_coord.z * b.m_coord.z) );
end;



end.
